#ifdef HAVE_CONFIG_H
# include "config.h"
#endif


#include "basso_cu_device_kernels.h"
#include "stdio.h"


__device__
void printBitwiseMatrix(const basso_matrix_t *mat)
{
	//printf("row:%llu col:%llu\n",mat->rows,mat->cols);
	for(int i=0;i<mat->rows;i++){
		for(int j=0;j<mat->cols;j++){
			if(mat->majority == BASSO_COL_MAJ)
				printf("%d",((mat->data[j * mat->elem_per_vect + i/basso_elem_bits] & BASSO_ONE_AT_IND(i)) != 0 ? 1:0));
			else
				printf("%d",((mat->data[i * mat->elem_per_vect + j/basso_elem_bits] & BASSO_ONE_AT_IND(j)) != 0 ? 1:0));
		}
		printf("\n");
	}
	printf("\n");
}

__global__ void
print_kernel(const basso_matrix_t mat)
{
	printBitwiseMatrix(&mat);
}

__inline__ __device__
long long ll_warp_sum_reduce(long long val,int reduce_size) {
#pragma unroll
	for (int mask = reduce_size/2; mask > 0; mask /= 2)
		val += __shfl_down(val, mask);
	return val;
}


__inline__ __device__
uint64_t ull_warp_sum_reduce(uint64_t val,int reduce_size) {
#pragma unroll
	for (int mask = reduce_size/2; mask > 0; mask /= 2)
		val += __shfl_down(val, mask);
	return val;
}

__inline__ __device__
uint64_t warp_Or_reduce(uint64_t val,int reduce_size) {
#pragma unroll
	for (int mask = reduce_size/2; mask > 0; mask /= 2)
		val |= __shfl_down(val, mask);
	return val;
}

__inline__ __device__ basso_elem_t get_element_cm_matrix(const basso_matrix_t matrix,const uint64_t col,const uint64_t element)
{
	return matrix.data[col * matrix.elem_per_vect + element];
}

__inline__ __device__ basso_elem_t get_element_rm_matrix(const basso_matrix_t matrix,const uint64_t row,const uint64_t element)
{
	return matrix.data[row * matrix.elem_per_vect + element];
}

__inline__ __device__ bool is_index_inside_cm_matrix(const basso_matrix_t matrix,const uint64_t col,const uint64_t element)
{
	return (col < matrix.cols && element < matrix.elem_per_vect);
}

__inline__ __device__ bool is_index_inside_rm_matrix(const basso_matrix_t matrix,const uint64_t row,const uint64_t element)
{
	return (row < matrix.rows && element < matrix.elem_per_vect);
}

static __inline__ __device__
uint8_t
get_element_covered_1s_score(	const basso_elem_t 	&elem_A,
								const basso_elem_t 	&elem_Covered,
								const basso_elem_t 	&elem_candidate)
{
	// find the newly covered 1s
	basso_elem_t newlyCovered;
	newlyCovered  = ~(elem_Covered);
	newlyCovered &=   elem_A;
	newlyCovered &=   elem_candidate;

	return __popcll((unsigned long long)newlyCovered);
}

static __inline__ __device__
uint8_t
get_element_covered_0s_score(	const basso_elem_t 	&elem_A,
								const basso_elem_t 	&elem_Covered,
								const basso_elem_t 	&elem_candidate)
{
	// find the newly covered 0s
	basso_elem_t newlyOvercovered;
	newlyOvercovered  = elem_A;
	newlyOvercovered |= elem_Covered;
	newlyOvercovered  = ~newlyOvercovered;
	newlyOvercovered &= elem_candidate;

	return __popcll((unsigned long long)newlyOvercovered);
}

//static __inline__ __device__
//long long
//get_element_covered_score(	const basso_elem_t 	&elem_A,
//							const basso_elem_t 	&elem_Covered,
//							const basso_elem_t 	&elem_candidate,
//							const uint64_t		&covering_1s_weight)
//{
//	uint8_t covered_1s = 0;
//	uint8_t covered_0s = 0;
//
//	// find the newly covered 1s
//	basso_elem_t newlyCovered;
//	newlyCovered  = ~(elem_Covered);
//	newlyCovered &=   elem_A;
//	newlyCovered &=   elem_candidate;
//
//	// find the newly covered 0s
//	basso_elem_t newlyOvercovered;
//	newlyOvercovered  = elem_A;
//	newlyOvercovered |= elem_Covered;
//	newlyOvercovered  = ~newlyOvercovered;
//	newlyOvercovered &= elem_candidate;
//
//	covered_1s 	+= __popcll(newlyCovered);
//	covered_0s	+= __popcll(newlyOvercovered);
//
//	return (covering_1s_weight * covered_1s - covered_0s);
//}



__global__ void
basso_cu_get_association_matrix_kernel(const basso_matrix_t A1,const basso_matrix_t A2,const uint64_t C_row_offset,const double thresholds[],uint64_t thresholds_len,basso_matrix_t C)
{
	uint64_t A1_row_offset = blockIdx.y * BASSO_CU_BLOCKSIZE;
	uint64_t A2_row_offset = blockIdx.x * BASSO_CU_BLOCKSIZE;

	__shared__ basso_elem_t s_A1[BASSO_CU_BLOCKSIZE][BASSO_CU_BLOCKSIZE]; // Maybe be a problem when use basso_elem_t
	__shared__ basso_elem_t s_A2[BASSO_CU_BLOCKSIZE][BASSO_CU_BLOCKSIZE];

	int row = threadIdx.y;
	int col = threadIdx.x;

	uint64_t inner_product = 0;
	uint64_t ones_per_row = 0;
	for(uint64_t element_offset = 0 ; element_offset < A1.elem_per_vect ; element_offset += BASSO_CU_BLOCKSIZE)
	{
		s_A1[row][col] = (is_index_inside_rm_matrix(A1,A1_row_offset + row,element_offset + col) ? get_element_rm_matrix(A1,A1_row_offset + row,element_offset + col) : 0);
		s_A2[row][col] = (is_index_inside_rm_matrix(A2,A2_row_offset + row,element_offset + col) ? get_element_rm_matrix(A2,A2_row_offset + row,element_offset + col) : 0);

		__syncthreads();

		for(int l=0 ; l < BASSO_CU_BLOCKSIZE ; l++)
		{
			int k = (col + l) % BASSO_CU_BLOCKSIZE; //This mapping removes bank-conflict in shared memory
			ones_per_row += __popcll((unsigned long long) s_A2[row][k]);
			inner_product += __popcll((unsigned long long) s_A1[col][k] & s_A2[row][k]);
		}

		__syncthreads();
	}

	for(int i=0 ; i < thresholds_len ; i++)
	{
		double threshold = thresholds[i];
		uint64_t element_set_bits = 0;
		if(inner_product > threshold * ones_per_row)
			element_set_bits =  BASSO_ONE_AT_IND(A1_row_offset + col);

		element_set_bits = warp_Or_reduce(element_set_bits,BASSO_CU_BLOCKSIZE);

		if(col == 0 && (A2_row_offset * thresholds_len + row * thresholds_len + i) < C.cols && (C_row_offset + A1_row_offset) / basso_elem_bits < C.elem_per_vect)
			atomicOr((unsigned long long *)C.data + (A2_row_offset * thresholds_len + row * thresholds_len + i) * C.elem_per_vect + (C_row_offset + A1_row_offset) / basso_elem_bits,element_set_bits);

		__syncthreads();
	}
}

__global__ void
basso_cu_get_covered_scores_kernel(const basso_matrix_t A,const basso_matrix_t C,const basso_matrix_t Covered, const uint64_t covering_1s_weight,const uint64_t covering_0s_weight,uint64_t *scores)
{
	uint64_t C_col_offset = blockIdx.x * BASSO_CU_BLOCKSIZE;

	__shared__ basso_elem_t s_A[BASSO_CU_BLOCKSIZE][BASSO_CU_BLOCKSIZE];
	__shared__ basso_elem_t s_C[BASSO_CU_BLOCKSIZE][BASSO_CU_BLOCKSIZE];
	__shared__ basso_elem_t s_Covered[BASSO_CU_BLOCKSIZE][BASSO_CU_BLOCKSIZE];

	int row = threadIdx.x;
	int col = threadIdx.y;

	uint64_t candidate_score = 0;
	for(uint64_t A_col_offset = 0 ; A_col_offset < A.cols ; A_col_offset += BASSO_CU_BLOCKSIZE)
	{
		uint64_t columnScore = 0;
		uint64_t columnOverScore = 0;

		for(uint64_t element_offset = 0 ; element_offset < A.elem_per_vect ; element_offset += BASSO_CU_BLOCKSIZE)
		{
			s_A[col][row] = (is_index_inside_cm_matrix(A, A_col_offset + col,element_offset + row) ? get_element_cm_matrix(A,A_col_offset + col,element_offset + row) : 0);
			s_Covered[col][row] = (is_index_inside_cm_matrix(Covered, A_col_offset + col, element_offset + row) ? get_element_cm_matrix(Covered, A_col_offset + col, element_offset + row) : 0);
			s_C[col][row] = (is_index_inside_cm_matrix(C, C_col_offset + col, element_offset + row) ? get_element_cm_matrix(C, C_col_offset + col, element_offset + row) : 0);

			__syncthreads();

			for(int l=0 ; l < BASSO_CU_BLOCKSIZE ; l++)
			{
				int k = (row + l) % BASSO_CU_BLOCKSIZE; //This mapping removes bank-conflict in shared memory

				columnScore 	+= get_element_covered_1s_score(s_A[row][k],s_Covered[row][k],s_C[col][k]);
				columnOverScore += get_element_covered_0s_score(s_A[row][k],s_Covered[row][k],s_C[col][k]);
			}

			__syncthreads();

		}

		const long long final_column_score = columnScore * covering_1s_weight - covering_0s_weight * columnOverScore;

		if(final_column_score > 0)
			candidate_score += final_column_score;
	}

	candidate_score = ull_warp_sum_reduce(candidate_score,BASSO_CU_BLOCKSIZE);

	if(row == 0)
		scores[C_col_offset + col] += candidate_score;
}


__global__ void
basso_cu_update_covered_scores_kernel(const basso_matrix_t A,const basso_matrix_t C,const basso_matrix_t Covered,const basso_elem_t *C_vector,const uint64_t covering_1s_weight,const uint64_t covering_0s_weight,long long *scores)
{
	uint64_t C_col_offset = blockIdx.x * BASSO_CU_BLOCKSIZE;

	__shared__ basso_elem_t s_A[BASSO_CU_BLOCKSIZE][BASSO_CU_BLOCKSIZE];
	__shared__ basso_elem_t s_C[BASSO_CU_BLOCKSIZE][BASSO_CU_BLOCKSIZE];
	__shared__ basso_elem_t s_Covered[BASSO_CU_BLOCKSIZE][BASSO_CU_BLOCKSIZE];
	__shared__ basso_elem_t s_Candidate[BASSO_CU_BLOCKSIZE];

	int row = threadIdx.x;
	int col = threadIdx.y;

	long long diff = 0;
	for(uint64_t A_col_offset = 0 ; A_col_offset < A.cols ; A_col_offset += BASSO_CU_BLOCKSIZE)
	{
		uint64_t prev_1s_score = 0 , prev_0s_score = 0;
		uint64_t new_1s_score = 0 , new_0s_score = 0;

		for(uint64_t element_offset = 0 ; element_offset < A.elem_per_vect ; element_offset += BASSO_CU_BLOCKSIZE)
		{
			s_A[col][row] = (is_index_inside_cm_matrix(A, A_col_offset + col,element_offset + row) ? get_element_cm_matrix(A,A_col_offset + col,element_offset + row) : 0);
			s_Covered[col][row] = (is_index_inside_cm_matrix(Covered, A_col_offset + col, element_offset + row) ? get_element_cm_matrix(Covered, A_col_offset + col, element_offset + row) : 0);
			s_C[col][row] = (is_index_inside_cm_matrix(C, C_col_offset + col, element_offset + row) ? get_element_cm_matrix(C, C_col_offset + col, element_offset + row) : 0);

			if(col == 0)
				s_Candidate[row] = (element_offset + row < C.elem_per_vect ? C_vector[element_offset + row] : 0);

			__syncthreads();

			for(int l=0 ; l < BASSO_CU_BLOCKSIZE ; l++)
			{
				int k = (row + l) % BASSO_CU_BLOCKSIZE; //This mapping removes bank-conflict in shared memory

				prev_1s_score += get_element_covered_1s_score(s_A[row][k],s_Covered[row][k],s_C[col][k]);
				prev_0s_score += get_element_covered_0s_score(s_A[row][k],s_Covered[row][k],s_C[col][k]);

				new_1s_score += get_element_covered_1s_score(s_A[row][k],s_Covered[row][k] | s_Candidate[k],s_C[col][k]);
				new_0s_score += get_element_covered_0s_score(s_A[row][k],s_Covered[row][k] | s_Candidate[k],s_C[col][k]);
			}

			__syncthreads();
		}

		long long prev_score 	= prev_1s_score * covering_1s_weight - prev_0s_score * covering_0s_weight;
		long long new_score 	= new_1s_score * covering_1s_weight - new_0s_score * covering_0s_weight;;

		if (prev_score > 0 && new_score > 0)
		{
			diff += new_score - prev_score;
		}
		/* Case 2: c was not used, but will now be used */
		else if (prev_score <= 0 && new_score > 0)
			diff += new_score;
		/* Case 3: c was used, but won’t be used anymore */
		else if (prev_score > 0 && new_score <= 0)
			diff -= prev_score;

	}

	diff = ll_warp_sum_reduce(diff,BASSO_CU_BLOCKSIZE);

	if(row == 0)
		scores[C_col_offset + col] += diff;
}


__global__ void
basso_cu_get_right_factor_vector_kernel(const basso_matrix_t A, const basso_matrix_t Covered, const basso_elem_t *C_vector, const uint64_t covering_1s_weight,const uint64_t covering_0s_weight, basso_elem_t *X_vector)
{
	uint64_t A_col_offset = blockIdx.x * BASSO_CU_BLOCKSIZE;

	__shared__ basso_elem_t s_C_vector[BASSO_CU_BLOCKSIZE * BASSO_CU_BLOCKSIZE];

	int row = threadIdx.x;
	int col = threadIdx.y;
	int tid = threadIdx.y * BASSO_CU_BLOCKSIZE + threadIdx.x;

	uint64_t columnScore = 0;
	uint64_t columnOverScore = 0;

	for(uint64_t C_vector_element_offset = 0 ; C_vector_element_offset < A.elem_per_vect ; C_vector_element_offset += BASSO_CU_BLOCKSIZE * BASSO_CU_BLOCKSIZE)
	{
		s_C_vector[tid] = (C_vector_element_offset + tid < A.elem_per_vect ? C_vector[C_vector_element_offset + tid] : 0);

		__syncthreads();

		for(uint64_t A_element_offset = 0 ; A_element_offset < BASSO_CU_BLOCKSIZE * BASSO_CU_BLOCKSIZE ; A_element_offset += BASSO_CU_BLOCKSIZE)
		{
			uint64_t A_element_idx = C_vector_element_offset + A_element_offset + row;

			const basso_elem_t Covered_element = (is_index_inside_cm_matrix(Covered, A_col_offset + col,A_element_idx) ? get_element_cm_matrix(Covered, A_col_offset + col,A_element_idx) : 0);
			const basso_elem_t A_element = (is_index_inside_cm_matrix(A, A_col_offset + col,A_element_idx) ? get_element_cm_matrix(A, A_col_offset + col,A_element_idx) : 0);
			const basso_elem_t C_vector_element = s_C_vector[A_element_offset + row];

			columnScore 	+= get_element_covered_1s_score(A_element,Covered_element,C_vector_element);
			columnOverScore += get_element_covered_0s_score(A_element,Covered_element,C_vector_element);
		}

		__syncthreads();
	}

	columnScore = ull_warp_sum_reduce(columnScore,BASSO_CU_BLOCKSIZE);
	columnOverScore = ull_warp_sum_reduce(columnOverScore,BASSO_CU_BLOCKSIZE);

	const long long final_score = columnScore * covering_1s_weight - columnOverScore * covering_0s_weight;

	if(row == 0 && A_col_offset + col < A.cols && final_score > 0)
		atomicOr((unsigned long long *) X_vector + (A_col_offset + col) / basso_elem_bits , BASSO_ONE_AT_IND(A_col_offset + col));
}


__global__ void
basso_cu_update_covered_matrix_kernel(const basso_elem_t *C_vector,const basso_elem_t *X_vector,basso_matrix_t Covered)
{
	uint64_t Covered_col_offset = blockIdx.x * BASSO_CU_BLOCKSIZE;

	__shared__ basso_elem_t s_C_vector[BASSO_CU_BLOCKSIZE * BASSO_CU_BLOCKSIZE];

	int row = threadIdx.x;
	int col = threadIdx.y;
	int tid = threadIdx.y * BASSO_CU_BLOCKSIZE + threadIdx.x;

	for(uint64_t C_vector_element_offset = 0 ; C_vector_element_offset < Covered.elem_per_vect ; C_vector_element_offset += BASSO_CU_BLOCKSIZE * BASSO_CU_BLOCKSIZE)
	{
		s_C_vector[tid] = (C_vector_element_offset + tid < Covered.elem_per_vect ? C_vector[C_vector_element_offset + tid] : 0);

		__syncthreads();

		for(uint64_t s_C_vector_element_idx = 0 ; s_C_vector_element_idx < BASSO_CU_BLOCKSIZE * BASSO_CU_BLOCKSIZE ; s_C_vector_element_idx += BASSO_CU_BLOCKSIZE)
		{
			uint64_t Covered_element_idx = C_vector_element_offset + s_C_vector_element_idx + row;
			if(Covered_element_idx >= Covered.elem_per_vect)
				break;

			if(Covered_col_offset + col < Covered.cols && (X_vector[(Covered_col_offset + col)/basso_elem_bits ] & BASSO_ONE_AT_IND(Covered_col_offset + col)) != 0)
			{
				Covered.data[(Covered_col_offset + col) * Covered.elem_per_vect + Covered_element_idx] |= s_C_vector[s_C_vector_element_idx + row];
			}
		}

		__syncthreads();
	}
}

__global__ void
basso_cu_update_covered_matrix_without_nth_factor_kernel(const basso_matrix_t B,basso_matrix_t covered_vector)
{
	uint64_t col_offset 	= blockIdx.x * BASSO_CU_BLOCKSIZE;
	uint64_t elem_offset 	= blockIdx.y * BASSO_CU_BLOCKSIZE;

	__shared__ basso_elem_t s_B[BASSO_CU_BLOCKSIZE][BASSO_CU_BLOCKSIZE];

	int col = threadIdx.y;
	int elem = threadIdx.x;

	s_B[col][elem] = (is_index_inside_cm_matrix(B, col_offset + col,elem_offset + elem) ? get_element_cm_matrix(B, col_offset + col,elem_offset + elem) : 0);

	__syncthreads();

	//We only need first warp and we release other warps
	if(threadIdx.y == 0)
	{
		basso_elem_t elem_OR = 0;
		for(int k=0 ; k < BASSO_CU_BLOCKSIZE ; k++)
		{
			elem_OR |= s_B[k][threadIdx.x];
		}

		if(elem_offset + threadIdx.x < B.elem_per_vect)
			atomicOr((unsigned long long *) covered_vector.data + elem_offset + threadIdx.x, elem_OR);
	}
}



