#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "basso_cu.h"
#include "../bassopriv.h"



#ifndef BASSO_ERRMSG
#  define BASSO_ERRMSG(msg) fprintf(stderr, "%s (%s:%d): %s\n", __func__, __FILE__, __LINE__, (msg))
#endif

basso_cu_block make_block(const uint64_t rows,const uint64_t cols){
	basso_cu_block ret;
	ret.rows = rows;
	ret.cols = cols;
	return ret;
};


int
basso_cu_m_block_copy(basso_matrix_t *dest, const basso_cu_block dest_offset, const basso_matrix_t *src, const basso_cu_block src_offset, const basso_cu_block size, const cudaStream_t stream ,enum cudaMemcpyKind kind)
{
	if(src->majority != dest->majority)
	{
		BASSO_ERRMSG("The majority of the source and destination matrices are not the same");
		return 1;
	}

	if(dest_offset.cols >= dest->cols || dest_offset.rows >= dest->rows || src_offset.cols >= src->cols || src_offset.rows >= src->rows)
	{
		BASSO_ERRMSG("The starting copy offsets is out of the bounds of source or destination matrix");
		return 1;
	}

	if(src_offset.rows + size.rows > src->rows || src_offset.cols + size.cols > src->cols || dest_offset.rows + size.rows > dest->rows || dest_offset.cols + size.cols > dest->cols)
	{
		BASSO_ERRMSG("The bounds of copy area (calculated by the offsets) are out of the bounds of source or destination matrix");
		return 1;
	}

	basso_majority_t majority = src->majority;

	if(majority == BASSO_ROW_MAJ)
	{

		uint64_t src_element_offset = (uint64_t)floorf((float)src_offset.cols/basso_elem_bits);//Move the offset to the beginning of basso_elem_t on source
		uint64_t dest_element_offset = (uint64_t)floorf((float)dest_offset.cols/basso_elem_bits);//Move the offset to the beginning of basso_elem_t on host
		uint64_t element_per_vect = (uint64_t)ceilf((float)size.cols/basso_elem_bits);//Copy size based on the number of basso_elem_t

		if(element_per_vect == src->elem_per_vect && element_per_vect == dest->elem_per_vect)
		{
			CUDA_CHECK_ERROR(cudaMemcpyAsync(	dest->data + dest_offset.rows * dest->elem_per_vect,
												src->data + src_offset.rows * src->elem_per_vect,
												size.rows * element_per_vect * sizeof(basso_elem_t),
												kind,
												stream));
		}
		else
		{
			fprintf(stderr, "%s (%s:%d): Warning: Many small memcpy\n", __func__, __FILE__, __LINE__);
			for(uint64_t i=0 ; i < size.rows ; i++)
			{
				CUDA_CHECK_ERROR(cudaMemcpyAsync(	dest->data + (dest_offset.rows + i) * dest->elem_per_vect + dest_element_offset,
													src->data + (src_offset.rows + i) * src->elem_per_vect  + src_element_offset,
													element_per_vect * sizeof(basso_elem_t),
													kind,
													stream));
			}
		}

		return 0;
	}
	else
	{

		uint64_t src_element_offset = (uint64_t)floorf((float)src_offset.rows/basso_elem_bits);
		uint64_t dest_element_offset = (uint64_t)floorf((float)dest_offset.rows/basso_elem_bits);
		uint64_t element_per_vect = (uint64_t)ceilf((float)size.rows/basso_elem_bits);

		if(element_per_vect == src->elem_per_vect && element_per_vect == dest->elem_per_vect)
		{
			CUDA_CHECK_ERROR(cudaMemcpyAsync(	dest->data + dest_offset.cols * dest->elem_per_vect,
												src->data + src_offset.cols * src->elem_per_vect,
												size.cols * element_per_vect * sizeof(basso_elem_t),
												kind,
												stream));
		}
		else
		{
			fprintf(stderr, "%s (%s:%d): Warning: Many small memcpy\n", __func__, __FILE__, __LINE__);
			for(uint64_t i=0 ; i < size.cols ; i++)
			{
				CUDA_CHECK_ERROR(cudaMemcpyAsync(	dest->data + (dest_offset.cols + i) * dest->elem_per_vect + dest_element_offset,
													src->data + (src_offset.cols + i) * src->elem_per_vect  + src_element_offset,
													element_per_vect * sizeof(basso_elem_t),
													kind,
													stream));
			}
		}

		return 0;
	}
}



int
basso_cu_m_alloc(const uint64_t rows, const uint64_t cols, const basso_majority_t majority,const basso_cu_memory_scope scope,basso_matrix_t ** matrix)
{
	basso_matrix_t *mat;
	if(cudaMallocHost((void**) &mat,sizeof(basso_matrix_t)) != cudaSuccess)
	{
		BASSO_ERRMSG("Failed to allocate pin memory for the matrix on host");
		return 1;
	}

	*matrix = mat;

	mat->rows = rows;
	mat->cols = cols;
	mat->majority = majority;

	uint64_t data_size;
	if (majority == BASSO_ROW_MAJ) {
		mat->elem_per_vect = elem_per_vect(cols);
		data_size = rows * mat->elem_per_vect * sizeof(basso_elem_t);

	} else {
		mat->elem_per_vect = elem_per_vect(rows);
		data_size = cols * mat->elem_per_vect * sizeof(basso_elem_t);
	}

	if(scope == BASSO_CU_ON_DEVICE)
	{
		if(cudaMalloc((void**)&(mat->data),data_size) != cudaSuccess)
		{
			BASSO_ERRMSG("Failed to allocate memory for the data field of the matrix on device");
			CUDA_CHECK_ERROR(cudaFreeHost(mat));
			return 1;
		}

		if(cudaMemset(mat->data,0,data_size) != cudaSuccess)
		{
			BASSO_ERRMSG("Failed to initialize memory for the data field");
			CUDA_CHECK_ERROR(cudaFree(mat->data));
			CUDA_CHECK_ERROR(cudaFreeHost(mat));
			return 1;
		}

		return 0;
	}
	else
	{
		if(cudaMallocHost((void**)&(mat->data),data_size) != cudaSuccess)
		{
			BASSO_ERRMSG("Failed to allocate pin memory for the data field of the matrix on host");
			CUDA_CHECK_ERROR(cudaFreeHost(mat));
			return 1;
		}

		memset(mat->data,0,data_size);

		return 0;
	}
}

int
basso_cu_m_device_set_elements_to_zero(basso_matrix_t* matrix,const cudaStream_t stream)
{
	uint64_t data_size = ((matrix->majority == BASSO_ROW_MAJ ? matrix->rows : matrix->cols) * matrix->elem_per_vect * sizeof(basso_elem_t));
	if(cudaMemsetAsync(matrix->data,0,data_size,stream) != cudaSuccess)
	{
		BASSO_ERRMSG("Failed to set the data field of the matrix on device to zero");
		return 1;
	}
	return 0;
}


int
basso_cu_m_free(basso_matrix_t* matrix,basso_cu_memory_scope scope)
{
	int status = 0;

	if(scope == BASSO_CU_ON_DEVICE)
	{
		if(cudaFree(matrix->data) != cudaSuccess)
		{
			BASSO_ERRMSG("Failed to free the memory of the data field of the matrix on device");
			status = 1;
		}
	}
	else
	{
		if(cudaFreeHost(matrix->data) != cudaSuccess)
		{
			BASSO_ERRMSG("Failed to free the pin memory of the data field of the matrix on host");
			status = 1;
		}
	}

	if(cudaFreeHost(matrix) != cudaSuccess)
	{
		BASSO_ERRMSG("Failed to free the pin memory of the matrix on host");
		status = 1;
	}

	return status;
}


/*

void 
basso_cu_m_cpy_host_to_device(basso_matrix_t **dest,const basso_matrix_t *src)
{
	basso_matrix_t *h_matrix = (basso_matrix_t*)malloc(sizeof(basso_matrix_t));
	basso_matrix_t *d_matrix;

	h_matrix->majority = src->majority;
	h_matrix->rows = src->rows;
	h_matrix->cols = src->cols;
	h_matrix->elem_per_vect = src->elem_per_vect;


	size_t data_size = (src->majority == BASSO_ROW_MAJ ? src->rows : src->cols) * (src->elem_per_vect) * sizeof(basso_elem_t);
	cudaMalloc((void**)&(h_matrix->data) , data_size);
	cudaMemcpy(h_matrix->data , src->data , data_size , cudaMemcpyHostToDevice);


	cudaMalloc((void**)&d_matrix , sizeof(basso_matrix_t));
	cudaMemcpy(d_matrix , h_matrix , sizeof(basso_matrix_t) , cudaMemcpyHostToDevice);

	free(h_matrix);

	*dest = d_matrix;
}

void 
basso_cu_m_cpy_device_to_host(basso_matrix_t **dest,const basso_matrix_t *src)
{
	basso_matrix_t *h_matrix = (basso_matrix_t*)malloc(sizeof(basso_matrix_t));
	cudaMemcpy(h_matrix , src , sizeof(basso_matrix_t) , cudaMemcpyDeviceToHost);


	basso_elem_t *d_data = h_matrix->data;
	size_t data_size = (h_matrix->majority == BASSO_ROW_MAJ ? h_matrix->rows : h_matrix->cols) * (h_matrix->elem_per_vect) * sizeof(basso_elem_t);
	h_matrix->data = (basso_elem_t *)malloc(data_size);
	cudaMemcpy(h_matrix->data,d_data,data_size,cudaMemcpyDeviceToHost);

	*dest = h_matrix;
}



basso_matrix_t *
basso_cu_m_alloc(const size_t rows, const size_t cols, const basso_majority_t majority)
{
	basso_matrix_t *h_A_meta = (basso_matrix_t *)malloc(sizeof(basso_matrix_t));

	h_A_meta->rows = rows;
	h_A_meta->cols = cols;
	h_A_meta->majority = majority;

	size_t data_size;
	if (majority == BASSO_ROW_MAJ) {
		h_A_meta->elem_per_vect = elem_per_vect(cols);
		data_size = rows * h_A_meta->elem_per_vect * sizeof(basso_elem_t);

	} else {
		h_A_meta->elem_per_vect = elem_per_vect(rows);
		data_size = cols * h_A_meta->elem_per_vect * sizeof(basso_elem_t);
	}

	if(cudaMalloc((void**)&(h_A_meta->data),data_size) != cudaSuccess)
		return NULL;

	if(cudaMemset(h_A_meta->data,0,data_size) != cudaSuccess)
	{
		cudaFree(h_A_meta->data);
		return NULL;
	}

	basso_matrix_t *d_A;
	if(cudaMalloc((void**)&d_A , sizeof(basso_matrix_t)) == cudaSuccess)
	{
		if(cudaMemcpy(d_A,h_A_meta,sizeof(basso_matrix_t),cudaMemcpyHostToDevice) == cudaSuccess)
			return d_A;
		else
		{
			cudaFree(h_A_meta->data);
			cudaFree(d_A);
			return NULL;
		}
	}
	else
	{
		cudaFree(h_A_meta->data);
		return NULL;
	}

}



void
basso_cu_m_free(basso_matrix_t* d_matrix)
{
	basso_matrix_t* h_matrix_meta = (basso_matrix_t*)malloc(sizeof(basso_matrix_t));
	cudaMemcpy(h_matrix_meta ,d_matrix,sizeof(basso_matrix_t),cudaMemcpyDeviceToHost);

	cudaFree(h_matrix_meta->data);
	cudaFree(d_matrix);
	free(h_matrix_meta);
}

basso_matrix_t
basso_cu_m_retrieve_info(const basso_matrix_t* d_matrix)
{
	basso_matrix_t info;
	cudaMemcpy(&info,d_matrix,sizeof(basso_matrix_t),cudaMemcpyDeviceToHost);
	return info;
}


basso_matrix_t
basso_m_sub_rows(const basso_matrix_t* mat,const uint64_t start_row,const uint64_t num_rows)
{
	basso_matrix_t sub_mat;
	if(start_row >= mat->rows || start_row + num_rows > mat->rows)
	{
		BASSO_ERRMSG("Index out of range");
		return sub_mat;
	}

	if(mat->majority != BASSO_ROW_MAJ)
	{
		BASSO_ERRMSG("The function only works for row-major matrices");
		return sub_mat;
	}

	sub_mat.cols = mat->cols;
	sub_mat.rows = num_rows;
	sub_mat.elem_per_vect = mat->elem_per_vect;
	sub_mat.majority = mat->majority;
	sub_mat.data = (mat->data + start_row * mat->elem_per_vect);

	return sub_mat;
}
*/

